#!/usr/bin/perl
###################################################################
# For complaints concerning this sloppy code e-mail to the        #
# culprit : suh@info.ucl.ac.be.                                   #
# DATE : 07/2004.                                                 #
###################################################################
# $ARGV[0] : GEANT RIB                                            #
# $ARGV[1] : Directory with BGP updates                           #
# $ARGV[2] : IP of external peer whose peering must go down       #
# $ARGV[3] : IP of GEANT router having the peering                #
# $ARGV[4] : Hour when the peering is to be considered down       #
# $ARGV[5] : Number of hours that the peering must be held down   #
# $ARGV[6] : Number of hours of BGP updates to consider           #
###################################################################

# The purpose of this script is to read the BGP data from GEANT and
# simulate peerings down (and up) events directly inside the eBGP
# updates. For that we record maintain the RIB of each external
# peer of GEANT and then at some time we announce withdraw messages
# for all routes of the external peer, then we reannounce the last
# state of the RIB of the peer before the peering down event.
# 
# You need to specify the IP of the external peer whose peering is to
# be brought down, the time that this down period must last, and the

use IPC::Open2;
use IO::Handle;
use File::Find;
use Symbol;
use Date::Manip;
use Net::Patricia;# structure for BGP routing table

#########################
# Variables declaration #
#########################

# BGP routing table of external peer to be played with.
%bgp;
$bgp = new Net::Patricia;

$BASETIME = 0;
# The beginning timestamp we need to take into account is the one of the RIB.
# So we convert the dates we see in the files into UNIX time.
@FILENAME = split(/\//,$ARGV[0]);
@DATE = split(/\./,$FILENAME[scalar(@FILENAME)-1]);
# Last element of @DATE should be "gz", previous the hour and minutes, and
# the previous year month and day.
$month = substr($DATE[scalar(@DATE)-3],4,2);
$day = substr($DATE[scalar(@DATE)-3],6,2);
$year = substr($DATE[scalar(@DATE)-3],0,4);
$hour = substr($DATE[scalar(@DATE)-2],0,2);
$minutes = substr($DATE[scalar(@DATE)-2],2,2);
# Base UNIX time is as follows.
$BASETIME = &Date_SecsSince1970($month,$day,$year,$hour,$minutes,0);

$MAXTIME = $ARGV[6];

################################################################################################
# Reading the BGP RIB.
################################################################################################
# Reading the BGP RIB to know which peer announces which route towards the prefixes for which
# we have some traffic.
print "Reading BGP RIB\n";
read_BGP($ARGV[0]);

# Creating directory for modified updates. 
system("mkdir updates\_$ARGV[2]\_$ARGV[4]\_$ARGV[5]");

################################################################################################
# Reading BGP updates.
################################################################################################
foreach $time (0..($MAXTIME-1)){
	load_updates($time);
};


################################################################################################
#                                            ###############                                   #
#                                            # Subroutines #                                   #
#                                            ###############                                   #
################################################################################################


############################
# Reads BGP RIB
############################
# This subroutine reads the BGP RIB and maintains all the routes that need to be announced
# through the virtual peers.

sub read_BGP {
	my $pid;
	if (not defined($pid = open(IN, "-|"))) {
		die "can't fork: $!";
	}
	if ($pid) {
		# parent process - do nothing
	} 
	else {
		# child process
		system("zcat $_[0] | route_btoa -m ");
		exit 0;
	};
	while (defined($CURRENTLINE=<IN>)){
		chop($CURRENTLINE);
		my @route = split(/\|/,$CURRENTLINE);
		# We record the route only if the BGP next hop is the one for which we simulate a peering down
		# event.
		if ($route[8] eq $ARGV[2]){
			# Put this route into the BGP routing table of the external peer.
			$bgp->add_string($route[5],"$CURRENTLINE");
		}
		else{
			# Do nothing.
			#print $CURRENTLINE;
		};
	};# end while
}


#####################
# Load BGP updates
#####################
# This subroutine takes as input the time (hour) for which the BGP updates need to be loaded
# and filters them to maintain only those that are relevant for the traffic.

sub load_updates {

	my $time = $_[0];
	if ($time > 0){
		$BASETIME+= 3600;
	};
	#print "Base Unix time is $BASETIME\n";
	# Now we can begin searching for update files that are within the right hour.
	# The right hour is between BASETIME and BASETIME+3600.
	%FILELIST = ();
	finddepth(\&filter_updates,$ARGV[1]);
	# So now we should have in FILELIST the list of update files to consider for this
	# time period. We sort the name of the files according to their UNIX timestamps.
	@updatefiles = sort {$a <=> $b} (keys %FILELIST);
	# We have the name of the BGP updates files to be considered. We take the name of the first
	# file and we open in output mode a similar file in the $ARGV[6] directory under the same
	# name.
	# Splitting file name
	my @FILENAME = split(/\//,$FILELIST{$updatefiles[0]});
	my $filestring = $FILENAME[scalar(@FILENAME)-1];
	my $pid;
	if (not defined($pid = open(OUT, "|-"))) {
		die "can't fork output pipe with gzip: $!";
	};
	if ($pid) {
		# parent process - do nothing
	} 
	else {
		# child process
		system("gzip -9 > updates\_$ARGV[2]\_$ARGV[4]\_$ARGV[5]\/$filestring");
		exit 0;
	};
	
	# If time is the one of the peer down event, climb the Patricia tree of the peer and announce
	# withdraw messages for all prefixes. If the peering down period has expired, we generate announce
	# messages for all prefixes of the RIB.
	if ($time == $ARGV[4]){
		print "Peering down event at time $time\n";
		$bgp->climb(\&generate_withdraw);
	};
	if($time == ($ARGV[4]+$ARGV[5])){
		print "Peering up event at time $time\n";
		$bgp->climb(\&generate_announce);
	};
	# Now we load these update files one at a time.
	foreach $file (@updatefiles){
		# Working on file $FILELIST{$file}.
		#print "Loading update file $FILELIST{$file}\n";
		# Opening the update file and filtering these updates since we only need
		# to care about those for which we have traffic.
		read_updates($FILELIST{$file},$time);
	};
	close(OUT);
}

############################################
# Generate withdraw for peering down event
############################################
# This subroutine generates a withdraw BGP message for each entry of the RIB of the external
# peer that undergoes a peering down event.

sub generate_withdraw {
	# Get the route for this prefix, and build the BGP message to be sent.
	my @route = split(/\|/,$_[0]);
	# To build the update, we only need the first six fields.
	print OUT "BGP4MP".'|'.$route[1].'|W|'.$route[3].'|'.$route[4].'|'.$route[5]."\n";
}

############################################
# Generate announce for peering up event
############################################
# This subroutine generates a announce BGP message for each entry of the RIB of the external
# peer that undergoes a peering up event.

sub generate_announce {
	# Easy, just get the route contained in the Patricia tree entry.
	# Get the route for this prefix, and build the BGP message to be sent.
	my @route = split(/\|/,$_[0]);
	print OUT "BGP4MP".'|'.$route[1].'|A|'.$route[3].'|'.$route[4].'|'.$route[5].'|'.$route[6].'|'.$route[7].
	'|'.$route[8].'|'.$route[9].'|'.$route[10].'|'.$route[11].'|NAG||'."\n";
}


#####################
# Read BGP updates
#####################
# This subroutine takes as input a file that contains BGP updates.

sub read_updates ($$) {
	my $file = $_[0];
	my $time = $_[1];
	#print "Reading updates file $file\n";
	my $pid;
	if (not defined($pid = open(IN, "-|"))) {
		die "can't fork: $!";
	}
	if ($pid) {
		# parent process - do nothing
	} 
	else {
		# child process
		system("zcat $_[0] | route_btoa -m ");
		exit 0;
	};
	while (defined($CURRENTLINE=<IN>)){
		chomp($CURRENTLINE);
		my @route = split(/\|/,$CURRENTLINE);
		if (($route[2] eq 'W')&&($route[3] eq $ARGV[3])){
			if (($time < $ARGV[4])||($time >= ($ARGV[4]+$ARGV[5]))){
				# We write this update as we are outside the peering down period.
				print OUT $CURRENTLINE."\n";
			};
			# Withdraw received. Update info in RIB and write message.
			if (defined($temp = $bgp->match_exact_string($route[5]))){
				$bgp->remove_string($route[5]);
			}
			else{
				# Trying to remove a prefix that is not in the RIB.
			};
		}
		elsif(($route[2] eq 'A')&&($route[8] eq $ARGV[2])){
			if (($time < $ARGV[4])||($time >= ($ARGV[4]+$ARGV[5]))){
				# We write this update as we are outside the peering down period.
				print OUT $CURRENTLINE."\n";
			};
			# Announce received from external peer. Update info in RIB and write message.
			if (defined($temp = $bgp->match_exact_string($route[5]))){
				$bgp->remove_string($route[5]);
			};
			$bgp->add_string($route[5],"$CURRENTLINE");
		}
		else{
			print OUT $CURRENTLINE."\n";
		};
	};# end while
	close(IN);
}

########################
# Filters update files
########################

sub filter_updates {
	# Checking that this is a real file.
	if (-f $File::Find::name){
		# Then checking that this is an update file with a correct name.
		my @FILENAME = split(/\//,$File::Find::name);
		my @DATE = split(/\./,$FILENAME[scalar(@FILENAME)-1]);
		# Last element of DATE should be 'gz', first should be 'updates.
		if (($DATE[scalar(@DATE)-1] eq 'gz')&&($DATE[0] eq 'updates')){
			# Converting this file's date info into Unix seconds.
			my $month = substr($DATE[scalar(@DATE)-3],4,2);
			my $day = substr($DATE[scalar(@DATE)-3],6,2);
			my $year = substr($DATE[scalar(@DATE)-3],0,4);
			my $hour = substr($DATE[scalar(@DATE)-2],0,2);
			my $minutes = substr($DATE[scalar(@DATE)-2],2,2);
			# Base UNIX time is as follows.
			$FILETIME = &Date_SecsSince1970($month,$day,$year,$hour,$minutes,0);
			# Now checking if the hour indicated in the file is within one hour
			# from the BASETIME.
			if (($FILETIME >= $BASETIME)&&(($FILETIME - $BASETIME) < 3600)){
				# Ok, file should be included in the list.
				$FILELIST{$FILETIME} = $File::Find::name;
				#print "Added file $FILELIST{$FILETIME} to the list of update files ($FILETIME  $BASETIME)\n";
			};
		};
	};
}
