#!/usr/bin/perl
###################################################################
# For complaints concerning this sloppy code e-mail to the        #
# culprit : suh@info.ucl.ac.be.                                   #
# DATE : 06/2004.                                                 #
###################################################################
# The purpose of this script is to simulate the optimization of   #
# the traffic of GEANT (the European Research Network). So it     #
# consists of the optimization of the choice of the egress router #
# to balance the traffic among the available exit points, given   #
# that we also want to minimize the IGP cost of the traffic.      #
# So this is a multi-objective problem, where we optimize the     #
# following objectives :                                          #
# - minimize the number of BGP changes (prefixes).                #
# - optimize the balance of the traffic among the exit points.    #
# - minimize the total IGP cost experienced by the traffic.       #
###################################################################
# This script implements a NSGA-II like algorithm in Perl that    #
# we use to sample a Pareto-front for three network objectives.   #
###################################################################
# Note that we solve this problem on-line, so we maintain the     #
# real BGP information to know the exact choice of all ingress    #
# routers of GEANT.
###################################################################
# $ARGV[0] : directory with traffic files for all ingresses       #
# $ARGV[1] : percentage of total traffic to be taken into account #
#            during the optimization (to know number of prefixes  #
#            that can be tweaked).                                #
# $ARGV[2] : cli file to be sent to CBGP                          #
# $ARGV[3] : GEANT RIB                                            #
# $ARGV[4] : Directory with BGP updates                           #
# $ARGV[5] : Number of time intervals to be optimized             #
# $ARGV[6] : Number of pairs to be used during the optimization   #
# $ARGV[7] : Lifetime of the tabu list entries in time intervals  #
# $ARGV[8] : Number of generations per interval                   #
###################################################################
# In this script, we change a subset of the ingresses to balance the traffic
# among the egress points while impacting as lightly as possible the hot-potato
# routing performed by each ingress router.

# This version of geant-online.pl relies on a tabu list to increase the
# efficiency of the optimization. At each time interval, we start with 
# the last best solution and we try to tweak additional ingress prefix
# pairs that are not contained in the tabu list.

use IPC::Open2;
use IO::Handle;
use File::Find;
use Symbol;
use CBGP 0.2;
use Date::Manip;

####################################################################################
# Variables and structures declaration, just trying to fool the reader that it was #
# cleanly designed ;-)                                                             #
####################################################################################

#############
# CONSTANTS #
#############
# Maximum population size, the actual population could vary but we define this bound just
# to ensure that the user is attentive to this parameter.
use constant MAXPOP => 1000;
# Maximum number of generations we compute during our search.
use constant MAXGEN => $ARGV[8];
# Maximum number of iterations of the BGP filter application tentatives over the whole
# population. This number means that the worst case is to iterate NUMITER times over
# the whole population in order to find improved solutions with an addition BGP filter.
use constant NUMITER => 1;
use constant MAXNONDOM => 100;
#########################
# Variables declaration #
#########################

%traffic;
%trafprefix;
%prefperegress;
$totcost;
%allrouteshash;
%peers;
%ingresses;
%exitrouter;
%igpcost;
%ingress2source;
%pairs;
%egressplusas2peer;
%bestroutes;
%rib_in;
%FILELIST;
%externalpeerings;
%common_tweaks;

@accepted;
@crowding_distance;
@crowd_dist_traffic;
@crowd_dist_igp;
@traffic_sorted;
@igp_sorted;
@source;
@orderedpairs;
@nondominated;
@packednondominated;
$accepted_pop_size;
# Hash containing for each time interval the solution chosen to be applied during the
# next hour.
@last_solution;

# Total traffic seen in the traffic stats.
$total = 0;

# List of known destination prefixes with traffic.
%prefixes;

# List of all prefixes by decreasing byte count.
@orderedprefixes;

# List of prefixes that can be tweaked by the optimization.
@tweakedprefixes;

# Number of prefixes that can be tweaked during the optimization, corresponds to $ARGV[1] percent
# of the total traffic.
$NUMPREFIXES = 0;

$LIFETIME = $ARGV[7];

# Number of time periods to be optimized (hours by default).
$MAXTIME = $ARGV[5];

$cbgp= CBGP->new("cbgp");
$cbgp->spawn;
print "spawned cbgp instance\n";


$BASETIME = 0;
@FILENAME = split(/\//,$ARGV[3]);
@DATE = split(/\./,$FILENAME[scalar(@FILENAME)-1]);
# Last element of @DATE should be "gz", previous the hour and minutes, and
# the previous year month and day.
$month = substr($DATE[scalar(@DATE)-3],4,2);
$day = substr($DATE[scalar(@DATE)-3],6,2);
$year = substr($DATE[scalar(@DATE)-3],0,4);
$hour = substr($DATE[scalar(@DATE)-2],0,2);
$minutes = substr($DATE[scalar(@DATE)-2],2,2);
# Base UNIX time is as follows.
$BASETIME = &Date_SecsSince1970($month,$day,$year,$hour,$minutes,0);

################################################################################################
# Reading the traffic files per ingress node.
################################################################################################
# The first thing to be done is reading the traffic files since it is based on this information that
# we choose which prefixes we are going to tweak during the optimization. So we compute the cumulative
# traffic per prefix and then take the largest prefixes that capture x % of the total traffic. 
print "reading traffic files\n";
find(\&read_traffic,$ARGV[0]);

print "Selecting largest prefixes to care about during the optimization\n";
@orderedprefixes = sort {$trafprefix{$b} <=> $trafprefix{$a}} (keys %trafprefix);
# Now computing the cumulative traffic percentage to know how many prefixes need to
# be taken into account to capture the largest fraction of the total traffic. We only
# need to care about some tens of percent of the total traffic.
my $temp = 0;
my $number = 0;
foreach my $pref (@orderedprefixes){
	# Computing cumulative traffic distribution and stopping as soon as we have
	# attained the percentage of the total traffic we need. The percentage required
	# is given as parameter ($ARGV[1]).
	$temp+=$trafprefix{$pref};
	$number++;
	# Testing is we are above the percentage.
	if ((($temp*100)/$total) >= $ARGV[1]){
		# Ok, got the traffic we need.
		print "To capture $ARGV[1] percent of the traffic we need to care about $number prefixes\n";
		$NUMPREFIXES = $number;
		last; # getting out from the loop.
	};
};

@tweakedprefixes = @orderedprefixes[0..($NUMPREFIXES-1)];
@orderedprefixes = ();

#############################
# Finding out GEANT routers.
#############################
# Reading the cli file to find out the GEANT routers.

print "Finding out GEANT routers based on cli file\n";
# Opening input and output pipe with cbgp.  
open(CLI,"<$ARGV[2]")|| die "cannot open input cli file: $!\n";
while (defined($CURRENTLINE=<CLI>)){
	chomp($CURRENTLINE);
	my @MYLIST = split(/[\s]+/,$CURRENTLINE);
	my @ADDRESS = split(/\./,$MYLIST[3]);
	if (($MYLIST[0] eq 'net')&&($MYLIST[1] eq 'add')&&($MYLIST[2] eq 'node')&&($ADDRESS[0] == 62)){
		# Found a new GEANT router.
		$ingresses{$MYLIST[3]}=1;
	};
	if (($MYLIST[0] eq '#')&&($MYLIST[1] eq 'peer')&&($MYLIST[2] eq 'router')){
		# External peer, record the internal GEANT IP + AS number to external IP mapping.
		$egressplusas2peer{"$MYLIST[3] $MYLIST[5]"} = $MYLIST[4];
		$externalpeerings{"$MYLIST[3] $MYLIST[4]"} = 1;
	};
};
close(CLI);

#####################################################
# Attributing each GEANT router to a traffic source.
#####################################################

$counter = 1;
foreach $in (keys %ingresses){
	# Attributing this ingress to the ith traffic source (if it exists).
	if (defined($source[$counter])){
		$ingress2source{$in} = $counter;
		$counter++;
	}
	else{
		# Error, don't have enough traffic sources.
		print "Not enough traffic sources to simulate ingress routers\n";
		exit(-1);
	};
};

################################################################################################
# Reading the BGP RIB.
################################################################################################
# Reading the BGP RIB to know which peer announces which route towards the prefixes for which
# we have some traffic.
$counter = 0;
read_BGP($ARGV[3]);
print "Read $counter routes for the prefixes with traffic\n";

################################################################################################
# Loading cli file
################################################################################################


# Sending cli file to cbgp. We read line by line and send each line to CBGP.
open(CLI,"<$ARGV[2]") or die "cannot open input cli file: $!\n";
while (defined($CURRENTLINE=<CLI>)){
	chomp($CURRENTLINE);
	$cbgp->send("$CURRENTLINE\n");
};
close(CLI) or die "cannot close input cli file: $!\n";

$cbgp->send("set autoflush on\n");

$cbgp->send("print \"CONFIGURATION OK\\n\"\n");
# Expect "CONFIGURATION OK"
$_= $cbgp->expect(1);
chomp;
if ($_ ne "CONFIGURATION OK") {
die "Beeeeeh !!!";
};

print "sent cli file to cbgp\n";

########################################################
# Replaying the BGP announcements made by peer routers.
########################################################
# We inject into CBGP the prefixes with traffic known by peers.
print "Injecting routes into CBGP to know best route for all ingresses and prefixes\n";
inject_routes();

# Now we know how by default BGP finds out the best route choice for each ingress.
# The real optimization can begin. We first compute the default objective value
# in terms of traffic balance and IGP cost. But for that we need to know the IGP
# cost from each ingress router towards each egress router.
print "Computing IGP costs between ingress and exit routers\n";
compute_IGP_cost();

# Now it's time to get into serious things. Let's begin the on-line optimization.
# For each time period, we start with the previous solution. This means that we
# take the tabu list that contains the ingresses and prefixes that are preferred
# compared other ingress and prefix pairs, advertize these changes into CBGP to
# get the default solution with the new traffic of the current time interval,
# apply the BGP updates for the current time interval to get the new BGP default
# solution and then we run the multi-objectives evolutionary search algorithm.
print "Let the optimization begin :)\n";
foreach my $time (0..($MAXTIME-1)){
	print "Time is $time\n";
	# We start by computing the default value of the traffic objectives when no tweaking of the routes
	# is performed.
	#print "Computing the default traffic objective for time $time\n";
	compute_default_obj($time);
	
	# So first we order the (ingress,prefix) pairs by decreasing byte count for the current time
	# interval, so that the evolutionary heuristic can work on the pairs that are most likely
	# to provide a gain in terms of the traffic balance while not increasing too much the
	# IGP cost.
	#print "Computing traffic and IGP cost per ingress prefix pair\n";
	compute_pairs($time);

	# Initializing the population at the default BGP solution. We do this because we want to compare
	# how stable one solution is to the next when both traffic and BGP changes.
	#print "Initializing population\n";
	init_pop($time);	

	# Note that we do not apply the solution found during the previous time interval but
	# start from scratch to prevent the old solution to force us to rely on tweaking that
	# are not optimal anymore. So what we do is to restart from scratch at each time
# 	# interval but at the end of the optimization we compute the delta between the solution
	# we just found and the previous one so that we advertize only changes with respect
	# to the old one in the network.
	#print "Getting into the optimization\n";
	# Adding one BGP tweaking at a time.
	my $generation = 1;
	while ($generation <= MAXGEN){
		# Freeing the data structures from the last iteration.
		@accepted = ();
		$accepted_pop_size = 0;
		@packednondominated = ();
		@nondominated = ();
		@crowding_distance = ();
		@crowd_dist_traffic = ();
		@crowd_dist_igp = ();
		@traffic_sorted = ();
		@igp_sorted = ();

		# Now getting into the job.
		my $iter = 0;
		$accepted_pop_size = 0;
		while (($accepted_pop_size < (MAXNONDOM))&&($iter < NUMITER)){
			print "Trying tweaking pairs to improve both traffic objectives (generation $generation)\n";
			try_tweaking($time);
			$iter++;
		};
		print "$accepted_pop_size individuals have been accepted during this iteration\n";
		# Finished tweaking individuals of the current population. We are not sure that the individuals accepted are non-dominated.
		# Actually some could be dominated. So we perform a non-domination check in the @nondominated array.
		#print "Checking for non-domination among the accepted individuals\n";
		nondomination_check();
		# Computing the crowding distance for the population. We shall use this crowding distance during
 		# the selection process.
		#print "Computing crowding distance\n";
		crowding_distance($time,$generation);
		#print "Comparing solutions found with the last one\n";
		if ($time > 0){
			compare_solutions($time,$generation);
		};
		# Freeing the @population array.
		@population = ();
		# Selecting and generating the new population.
		#print "Selecting individuals for next generation\n";
		select_generate_pop();
		$generation++;
	};
	# So before loading BGP updates, we look at the best solution found, i.e. the non-dominated solution that 
	# is best in terms of the IGP cost. So it is the non-dominated solution that is classified as first with
	# respect to the IGP cost.
	if ($time > 0){
		#print "Choosing best solution for this time interval\n";
		compare_solutions($time,$generation);
		choose_last($time);
	}
	else{
		# Time is zero, we cannot rely on choose_last since there is no
		# last solution for time -1. We just choose among the non-dominated
		# solutions of the last iteration the one with the smallest IGP cost.
		open(OUT,">>best-solutions")|| die "Could not open output file: $!\n";
		print OUT "$time $accepted[$packednondominated[$igp_sorted[0]][2]][0]:";
		my $elem = 1;
		my $counter = 0;
		while ($elem <= $accepted[$packednondominated[$igp_sorted[0]][2]][0]){
			my @tweak = split(/[\s]+/,$accepted[$packednondominated[$igp_sorted[0]][2]][$elem]);
			if ($tweak[4] == -1){
				print OUT " $accepted[$packednondominated[$igp_sorted[0]][2]][$elem]";
				# Ok, new entry. Initialize its value to $LIFETIME. We check if lifetime to be put is not 0
				# unless we cannot even add this tweak.
				if ($LIFETIME > 0){
					$last_solution[$time][$elem] = $tweak[0]." ".$tweak[1]." ".$tweak[2]." ".$tweak[3]." $LIFETIME";
					$counter++;
				}
				else{
					# Cannot add this tweak.
				};
			}
			else{
				# This should never happen since any tweak should be new hence have a lifetime set to -1.
				print "Tweak with abnormal lifetime found at time 0\n";
				exit(-1);
			};
			$elem++;
		};
		$last_solution[$time][0] = $counter;
		print OUT " $packednondominated[$igp_sorted[0]][6]\n";
		close(OUT)|| die "Could not close output file: $!\n";
	};
	# Now loading BGP updates for the next period.
	#print "Loading updates\n";
	load_updates($time); 
	# Re-setting the hashes that contain the RIB's.
	#%besroutes = ();
	#%rib_in = ();
	# Then update the ribs to take into account the changes in BGP.
	#print "Updating RIBs after updates injection\n";
	update_rib();
	%tobeupdated = ();
};

$cbgp->finalize;

################################################################################################
#                                            ###############                                   #
#                                            # Subroutines #                                   #
#                                            ###############                                   #
################################################################################################

############################
# Initializes traffic info
############################
# This subroutine takes as input the name of a directory containing the traffic files per ingress router.
# 

sub read_traffic {

	if (-f $File::Find::name){
		open(IN,"<$File::Find::name")|| die "cannot open input traffic file: $!\n";
		# The name of the file is source i, so we attribute source i to some ingress point (we don't care
		# which one actually).
		@NAME = split(/\//,$File::Find::name);
		$ingress = substr($NAME[scalar(@NAME)-1],6);
		# To tell that source i exists.
		$source[int($ingress)] = 1;
		#print "Reading traffic file of ingress $ingress\n";
		my $intraffic = 0;
		while (defined($CURRENTLINE=<IN>)){
			chop($CURRENTLINE);
			my @MYLIST = split(/[\s]+/,$CURRENTLINE);
			# Each line contains the traffic to be sent for each hour towards this prefix (first field)
			# during each hour of the trace.
#			foreach my $elem (0..(scalar(@MYLIST)-3)){
			foreach my $elem (0..($MAXTIME-1)){
				if (defined($MYLIST[$elem+2])){
					$traffic{$MYLIST[0]}{$MYLIST[1]}[$elem] = $MYLIST[$elem+2];
				}
				else{
					print "Not enough time intervals found in traffic\n";
					exit(-1);
				};
			};
			my $preftraf = 0;
			foreach my $elem (0..(scalar(@MYLIST)-3)){
				$traf = $traffic{$MYLIST[0]}{$MYLIST[1]}[$elem];
				$total += $traf;
				$intraffic += $traf;
				$preftraf += $traf;
			};
			# Recording total traffic for this prefix.
			my @LIST = split(/\./,$MYLIST[1]);
			# Checking if prefix looks like a prefix.
			if (scalar(@LIST) != 4){
				#print "This line has a problem : $CURRENTLINE\n";
			}
			else{
				$trafprefix{$MYLIST[1]} += $preftraf;
			};
			#print "traffic for prefix $MYLIST[1] and ingress $ingress is $preftraf\n";
		}
		close(IN);
		print "Total traffic read for ingress $ingress is: $intraffic\n";
	};
}

############################
# Reads BGP RIB
############################
# This subroutine reads the BGP RIB and maintains all the routes that need to be announced
# through the virtual peers.

sub read_BGP {
	my $pid;
	if (not defined($pid = open(IN, "-|"))) {
		die "can't fork: $!";
	}
	if ($pid) {
		# parent process - do nothing
	}
	else {
		# child process
		system("zcat $_[0] | route_btoa -m ");
		exit 0;
	};
	while (defined($CURRENTLINE=<IN>)){
		chop($CURRENTLINE);
		my @route = split(/\|/,$CURRENTLINE);
		# Accepting only routes that concern the destination prefixes.
		if (exists($trafprefix{$route[5]})){
			# So there is some traffic towards this destination prefix.
			# But still, we check that the next hop is really among those
			# that we expect (GlobalCrossing or Telia)
			my @ALIST = split(/\./,$route[8]);
			if (($ALIST[0] != 208)&&($ALIST[0] != 213)){
				# This should never happen, something is wrong in this route
				# since the next hop does not seem to be one of the peers
				# we expect.
				print "Oups \! Found next hop not among those we expect: $route[8]\n";
			}
			else{
				# Everything is fine, prefix has traffic and next hop is known.
				# We record this route since we'll inject it into CBGP when we'll
				# have seen all the routes towards those prefixes we're interested
				# in.		
				# Recording peer.
				$peers{$route[8]} = 1;
				# And also the exit GEANT router to reach this peer.
				$exitrouter{$route[3]} = 1;
				# Recording route information.
				$allrouteshash{$route[5]}{$route[8]} = $CURRENTLINE;
				$counter++;
				#print "$counter\n";
			};
		}
		else{
			#print $CURRENTLINE;
		};
	};# end while
}

############################
# Injects BGP routes
############################
# This subroutine injects the BGP routes for the prefixes with traffic. We announce all the routes
# for each prefix and record the RIB's all all ingress GEANT routers.

sub inject_routes {
	# For each prefix with traffic.
	my $router = '';
	foreach my $pref (keys %trafprefix){
		#print "Injecting routes for prefix $pref\n";
		# Now for each peer we check if it knows this prefix and if so announce it
		# through the corresponding virtual peer.
		foreach my $p (keys %peers){
			if (exists($allrouteshash{$pref}{$p})){
 				my @route = split(/\|/,$allrouteshash{$pref}{$p});
				# Checking a few things before injecting the route.
				if ($p ne $route[8]){
					# peer differs from the one that appears in the route
					print "Oups, peer differs from the one in the route: $p vs. $route[8]\n";
					print $allrouteshash{$pref}{$p},"\n";
					exit(-1);
				};
				# Injecting the route.
				$cbgp->send("bgp router $route[3] peer $p recv ".
				'"BGP4|0|A|'.$route[3].'|20965|'.$pref.'|'.$route[6].'|IGP|'.$p.'|'.$route[9].'|'.$route[10].'||NAG||"'."\n");
			};
		};
		# Letting CBGP run the simulation to let BGP converge.
		$cbgp->send("sim run\n");
		# Finally we ask to CBGP to dump the RIB of each ingress router of GEANT.
		foreach $in (keys %ingresses){
			# Asking best route for this ingress router.
			$cbgp->send("print \"\# $in RIB for prefix $pref\\n\" \n");
			$cbgp->send("bgp router $in show rib-in * $pref\n");
		};
		$cbgp->send("print \"done\\n\"\n"); 
		# Finished sending routes to CBGP for this prefix. Now we read from the input pipe the output
		# of CBGP.
		while ($_= $cbgp->expect(1)){
			chomp;
			m/^done/ and last;
			# Working on line we read.
			# If line begins with "#" then it contains the IP of the GEANT router
			# whose RIB we read. Next line then contains the best route to reach
			# prefix $pref.
			#print $_;
			my @temp = split(/[\s]+/,$_);
			if ($temp[0] eq '#'){
				# Read next line, record which router it is the RIB.
				$router = $temp[1];
				$prefix = $temp[5];
			}
			elsif ($temp[0] eq '*>'){
				# Ok, checking that prefix is the same in the RIB
				if ($temp[1] ne $prefix){
					print "Prefix in RIB-in ($temp[1]) is different from expected one: $prefix\n";
				};
				if (($router ne '')&&($temp[2] ne '')){
					$bestroutes{$router}{$pref} = "$temp[2] $temp[5]";
					#print "best egress for $router to reach $pref is $temp[2] $temp[5]\n";
				}
				else{
					print "Found void router in inject_routes ($router $temp[2])\n";
				};
			}
			elsif ($temp[0] eq '*'){
				if ($temp[1] ne $prefix){
					print "Prefix in RIB-in ($temp[1]) is different from expected one: $prefix\n";
				};
				my @toto = split(/\./,$temp[2]);
				if ($toto[0] eq '62'){
					if (($router ne '')&&($temp[2] ne '')){
						$rib_in{$router}{$pref}{$egressplusas2peer{"$temp[2] $temp[5]"}} = "$temp[2] $temp[5]";
						#print "egress for $router to reach $pref is ",$egressplusas2peer{"$temp[2] $temp[5]"},"\n"; 
					}
					else{
						print "Found void router in inject_routes ($router $temp[2])\n";
					};
				}
				else{
					if (($router ne '')&&($temp[2] ne '')){
						$rib_in{$router}{$pref}{$temp[2]} = "$temp[2] $temp[5]";
						#print "egress for $router to reach $pref is $temp[2]\n"; 
					}
					else{
						print "Found void router in inject_routes ($router $temp[2])\n";
					};
				};
			}
			else{
				print "Oups, this is a strange line received from cbgp: $_";
			};
		};
		# No more line to read from CBGP.
		#print "No more line to read from CBGP, going to next prefix\n";	
	};
}

############################
# Compute IGP costs
############################
# This subroutine computes the IGP cost from each ingress router towards each exit router.

sub compute_IGP_cost {

	# Opening input and output pipe with cbgp.  
	# Now we give to CBGP the commands to dump the best route to reach each egress router.
	foreach $in (keys %ingresses){
		# Asking the routing table of this ingress router.
		$cbgp->send("print \"\# $in RT\\n\" \n");
		$cbgp->send("net node $in show rt *\n");
		$cbgp->send("print \"done\\n\"\n");
		# Now reading output from CBGP.
		while ($_= $cbgp->expect(1)){
			chomp;
			m/^done/ and last;
			# If line begins with "#" then it contains the IP of the GEANT router
			# whose RIB we read. Next line then contains the best route to reach
			# prefix $pref.
			my @temp = split(/[\s]+/,$_);
			if ($temp[2] eq 'RT'){
				# Read next line, record which router it is the RIB.
				$router = $temp[1];
			}
			else{
				# So the first field contains destination prefix, so we need to
				# remove the /32 from the first field as well as check whether
				# this destination prefix concerns an exit router.
				my @exit = split(/\//,$temp[0]);
				if (exists($exitrouter{$exit[0]})){
					# Ok, this prefix corresponds to a known exit router,
					# so we record the IGP cost to reach it from this ingress.
					if (($router ne '')&&($temp[2] ne '')){
						$igpcost{$router}{$exit[0]} = $temp[2];
						#print "IGP cost for $router to reach egress $exit[0] is $temp[2]\n"; 
					}
					else{
						print "Found void router in compute_IGP_cost ($router $temp[2])\n";
					};
				};
			};
		};
	};	
}

####################################
# Compute default traffic objective
####################################
# This subroutine computes the traffic objectives for some time period before the optimization takes place.

sub compute_default_obj {

	my $time = $_[0];
	$totcost = 0;
	%trafperegress = ();
	foreach $pref (keys %trafprefix){
		# For each ingress point, compute the traffic times IGP cost product, as well as the
		# traffic per egress point.
		foreach $in (keys %ingresses){
			# Getting best exit point for ingress and prefix.
			my ($exit,$as) = split(/[\s]+/,$bestroutes{$in}{$pref});
			if ($exit ne ''){
				# Two cases, either the ingress router is directly connected to the peer,
				# in that case the IGP cost will be zero. In the case of a peer, then we directly
				# attribute the traffic to the egress router.
				my @temp = split(/\./,$exit);
				if (($temp[0] == 208)||($temp[0] == 213)){
					# Exit point is a peer, IGP cost is 0 so total cost does not change.
					# But traffic towards peer can be attributed.
					$trafperegress{$exit}+=$traffic{$ingress2source{$in}}{$pref}[$time];
				}
				else{
					# Exit point is internal router of GEANT, so we need to find the corresponding
					# peer router to attribute traffic.
						$trafperegress{$egressplusas2peer{"$exit $as"}}+=$traffic{$ingress2source{$in}}{$pref}[$time];
						# Exit point is internal router, get its IGP cost.
						$totcost += ($igpcost{$in}{$exit}*$traffic{$ingress2source{$in}}{$pref}[$time]);
				};
			}
			else{
				print "Found void exit router in compute_default_obj ($exit)\n";
				#print "route found in compute_default_obj $bestroutes{$in}{$pref}\n";
			};	
		};
	};

	# To improve speed we also apply the tweaks that are contained in the last solution on which the optimization
	# is going to rely instead of having to do that for each individual. Do that only if time is at least 1 unless
	# the last solution is empty...
	my $tweak = 1;
	if ($time > 0){
		while ($tweak < $last_solution[$time-1][0]){
			my ($in,$pref,$exit,$exitas,$ttl) = split(/[\s]+/,$last_solution[$time-1][$tweak]);
			my @tempexit = split(/\./,$exit);
			my ($default,$defaultas) = split(/[\s]+/,$bestroutes{$in}{$pref});
			my @tempdefault = split(/\./,$default);
			if (($default ne '')&&($exit ne '')){
				if ($tempdefault[0] ne '62'){
					# Default exit point is external. Remove traffic.
					$trafperegress{$default}-=$traffic{$ingress2source{$in}}{$pref}[$time];
					# Do not remove anything to the IGP cost since it is 0 for external peers.
				}
				else{
					# Default exit point is internal. Convert default exit point to peer.
					$trafperegress{$egressplusas2peer{"$default $defaultas"}}-=$traffic{$ingress2source{$in}}{$pref}[$time];
					# Remove contribution to IGP cost due to internal default exit point.
					$totcost-=($igpcost{$in}{$exit}*$traffic{$ingress2source{$in}}{$pref}[$time]);
				};	
				# Now adding traffic to new exit point.
				if ($tempexit[0] ne '62'){
					# New exit point is external. Add traffic contribution.
					$trafperegress{$exit}+=$traffic{$ingress2source{$in}}{$pref}[$time];
					# No IGP cost to add since it is 0 for external peers.
				}
				else{
					# New exit point is internal. Convert new exit point to peer.
					$trafperegress{$egressplusas2peer{"$exit $exitas"}}+=$traffic{$ingress2source{$in}}{$pref}[$time];
					# Add contribution to IGP cost.
					$totcost+=($igpcost{$in}{$exit}*$traffic{$ingress2source{$in}}{$pref}[$time]);
				};
			}
			else{
				print "Found void exit router in compute_default_obj ($default $exit)\n";
			};
			$tweak++;
		};
	};
	# Printing into the optimization output file the result. Traffic objective is the sum of squares of differences
	# between the ideal balance and the current one.
	open(OUT,">>objective-space-results")|| die "cannot open output file for optimization results: $!\n";
	my $tot = 0;
	foreach $elem (keys %trafperegress){
		$tot+=$trafperegress{$elem};
	};
	my $forbruno = '';
	foreach $elem (keys %trafperegress){
	print "$elem\n";
		$forbruno.= ($trafperegress{$elem}/$tot)." ";
	};
	my $delta = 0;
	# Amount of traffic if balance was perfect.
	my $balance = ($tot/scalar(keys %trafperegress));
	foreach my $elem (keys %trafperegress){
		$delta+= (abs($trafperegress{$elem}-$balance)/$balance)**1;
	};
	my $delta2 = 0;
	foreach my $elem (keys %trafperegress){
		$delta2+= (abs($trafperegress{$elem}-$balance)/$balance)**2;
	};
	my $max = 0;
	my $firstmax = 0;
	foreach my $elem (keys %trafperegress){
		if ($firstmax < $trafperegress{$elem}){
			$firstmax = $trafperegress{$elem};
		};
		if ($max < abs(($trafperegress{$elem}/$tot) - (1/scalar(keys %trafperegress)))){
			$max = abs(($trafperegress{$elem}/$tot) - (1/scalar(keys %trafperegress)));
			$best = $elem;
		};
	
	};
	# Computing the real objective in terms of $max.
	$max = 1 + ($trafperegress{$elem}/$tot) - (1/scalar(keys %trafperegress));
	print OUT "$time 0 -1 ".($tweak-1)." $delta2 1 $delta ".($firstmax/$tot)." $forbruno\n";
	close(OUT);
}

###################################################################
# Computing the IGP cost and traffic for each ingress prefix pair
###################################################################
# This subroutine computes for each ingress and prefix pair the traffic it has during
# the time period given in parameter, and sorts this list in decreasing byte count.

sub compute_pairs {

	my $time = $_[0];
	# Resetting the hash.
	%pairs = ();
	foreach my $pref (@tweakedprefixes){
		# Trying all ingress points and computing both traffic and IGP cost to reach the exit point.
		foreach my $in (keys %ingresses){
			# Finding out exit point used by ingress point to reach prefix.
			my ($exit,$as) = split(/[\s]+/,$bestroutes{$in}{$pref});
			if ($exit ne ''){
				# Before taking into account this pair, we check that it is not in the last individual
				# since it will never be try by the algorithm anyway (only if time is at least 1).
				my $elem = 1;
				my $seen = 0;
				if ($time >= 1){			
					while ($elem < $last_solution[$time][0]){
						my @temp = split(/[\s]+/,$last_solution[$time-1][$elem]);
						if ("$temp[0] $temp[1]" eq "$in $pref"){
							$seen = 1;
							last;
						};
						$elem++;
					};
				};
				if ($seen == 0){
					# Two cases, either the ingress router is directly connected to the peer,
					# in that case the IGP cost will be zero. In the case of a peer, then we directly
					# attribute the traffic to the egress router.
					my @temp = split(/\./,$exit);
					if (($temp[0] == 208)||($temp[0] == 213)){
						# Exit point is a peer, getting traffic.
						$pairs{"$in $pref"}[0] = $traffic{$ingress2source{$in}}{$pref}[$time];
						# IGP cost is 0 since we leave the AS directly. Still we put a cost of 1 since
						# we'll divide by the IGP cost later on.
						$pairs{"$in $pref"}[1] = O;
					}
					else{
						# Exit point is an internal router of GEANT, getting traffic.
						$pairs{"$in $pref"}[0] = $traffic{$ingress2source{$in}}{$pref}[$time];
						# IGP cost is 0 since we leave the AS directly.
						if (exists($igpcost{$in}{$exit})){
							$pairs{"$in $pref"}[1] = $igpcost{$in}{$exit};
						}
						else{
							print "Oups, do not have the IGP cost to reach prefix $pref from ingress $in (exit is 	$exit)\n"; 
						};
					};
				};
			}
			else{
				print "Found void exit router in compute_pairs ($exit)\n";
			};
		};
	};
	# Have traffic and IGP cost info for all ingress prefix pairs. Now we order them by increasing traffic times IGP cost.
	# We chose this order because it is the largest traffic sources that will not make the IGP cost increase too much that
	# we want to change.
#	my @orderedpairs = sort {($pairs{$b}[0]/$pairs{$b}[1]) <=> ($pairs{$a}[0]/$pairs{$a}[1])} (keys %pairs);
	@orderedpairs = sort {$pairs{$b}[0] <=> $pairs{$a}[0]} (keys %pairs);
	# Selecting only the largest pairs in terms of the traffic they will allow to move.
	# We take the $ARGV[6] largest ones.
	#print "Selected $ARGV[6] largest ingress prefix pairs to be optimized\n";
	if (scalar(@orderedpairs) > $ARGV[6]){
		@orderedpairs = @orderedpairs[0..($ARGV[6]-1)];
	}
	else{
		print "You tried to work on $ARGV[6] pairs while only ",scalar(@orderedpairs)," are available at time $time\n";
	};
	#my $counter = 1;
	#foreach my $elem (@orderedpairs){
	#	print "$counter: $pairs{$elem}[0] $pairs{$elem}[1]\n";
	#	$counter++;
	#};
}

###################################################################
# Computing the IGP cost and traffic for each ingress prefix pair
###################################################################
# This subroutine initializes the population to the solution chosen during
# the last time interval.

sub init_pop {
	my $time = $_[0];
	@population = ();
	if ($time == 0){
		foreach my $individual (0..(MAXPOP-1)){
			    $population[$individual][0] = 0;
		};
	}
	else{
		# Initializing to last solution since it is supposed to exist.
		foreach my $individual (0..(MAXPOP-1)){
			my $counter = 0;
			while ($counter <= $last_solution[$time-1][0]){
				$population[$individual][$counter] = $last_solution[$time-1][$counter];
				$counter++;
			};
		};
	};
}

###################################################################
# Computing the IGP cost and traffic for each ingress prefix pair
###################################################################
# This subroutine takes the current population and tries to add to each individual one more
# tweaking to improve both traffic objectives.

sub try_tweaking {
	my $time = $_[0];
	
	foreach my $individual (0..(MAXPOP-1)){
		# Selecting a random ingress prefix pair among those that can be tweaked.
		$tmppair = $orderedpairs[random_int_in(0,scalar(@orderedpairs)-1)];
		# Checking whether this filter has not already been applied on this individual.
		my $retries = 0;
		# Checking if pair is among current tweaks of the individual, but only starting at
		# tweak $last_solution[$time-1][0] since tweaks before that are old and cannot be
		# tried in the current interval.
		my $elem = 0;
		if ($time > 0){
			$elem = $last_solution[$time-1][0];
		}
		else{
			$elem = 1;
		};
		while (($elem <= $population[$individual][0])&&($retries < 5)){
			# Does this pair appear in the individual ?
			my @list = split(/[\s]+/,$population[$individual][$elem]);
			if ($tmppair eq "$list[0] $list[1]"){
				# Oups, need to try another pair. Marking as unchecked.
				$retries++;
				# Try another random pair.
				$tmppair = $orderedpairs[random_int_in(0,scalar(@orderedpairs)-1)];
				if ($time > 0){
					$elem = $last_solution[$time-1][0];
				}
				else{
					$elem = 1;
				};
			}
			else{
				$elem++;
			};
		};
		if ($elem > $population[$individual][0]){
			# Ok, pair can be tried.
			# Now we need to find out which exit point is best to improve balance while not increasing IGP
			# cost too much.
			
			# First copy traffic distribution and IGP cost into temporary variables.
			my %tmptrafdistri = ();
			foreach my $toto (keys %trafperegress){
				$tmptrafdistri{$toto} = $trafperegress{$toto};
			};
			my $tmpigpcost = $totcost;

			# Then apply the changes that are contained in this individual. But only those we added for this
			# time interval, not those from the last solution which have already been added during the computation
			# of the default value of the objectives.
			my $tweak;
			if ($time > 0){
				$tweak = $last_solution[$time-1][0];
			}
			else{
				$tweak = 1;
			};
			while ($tweak < $population[$individual][0]){
				my ($in,$pref,$exit,$exitas) = split(/[\s]+/,$population[$individual][$tweak]);
				my @tempexit = split(/\./,$exit);
				my ($default,$defaultas) = split(/[\s]+/,$bestroutes{$in}{$pref});
				my @tempdefault = split(/\./,$default);
				if ($tempdefault[0] ne '62'){
					# Default exit point is external. Remove traffic.
					$tmptrafdistri{$default}-=$traffic{$ingress2source{$in}}{$pref}[$time];
					# Do not remove anything to the IGP cost since it is 0 for external peers.
				}
				else{
					# Default exit point is internal. Convert default exit point to peer.
					$tmptrafdistri{$egressplusas2peer{"$default $defaultas"}}-=$traffic{$ingress2source{$in}}{$pref}[$time];
					# Remove contribution to IGP cost due to internal default exit point.
					$tmpigpcost-=($igpcost{$in}{$exit}*$traffic{$ingress2source{$in}}{$pref}[$time]);
				};	
				# Now adding traffic to new exit point.
				if ($tempexit[0] ne '62'){
					# New exit point is external. Add traffic contribution.
					$tmptrafdistri{$exit}+=$traffic{$ingress2source{$in}}{$pref}[$time];
					# No IGP cost to add since it is 0 for external peers.
				}
				else{
					# New exit point is internal. Convert new exit point to peer.
					$tmptrafdistri{$egressplusas2peer{"$exit $exitas"}}+=$traffic{$ingress2source{$in}}{$pref}[$time];
					# Add contribution to IGP cost.
					$tmpigpcost+=($igpcost{$in}{$exit}*$traffic{$ingress2source{$in}}{$pref}[$time]);
				};
				$tweak++;
			};
			# Computing objective value if we add this tweaking. It's quite simple, the traffic
			# objective is the max of the traffic, we try to balance it.
			my $oldmax = 0;
			foreach my $tutu (keys %tmptrafdistri){
				if ($tmptrafdistri{$tutu} > $oldmax){
					$oldmax = $tmptrafdistri{$tutu};
				};
			};
						
			# Finished applying changes contained in individual. Try to use exit points for this ingress and prefix
			# pair that improve the traffic balance.
			# Trying all possible exit points except the default one. So searching for the rib-in entries that
			# are in the rib-in hash.
			my ($in,$pref) = split(/[\s]+/,$tmppair);
			#print "Trying pair $in $pref\n";
			foreach my $exit (keys %peers){
			#print "Trying exit point $exit\n";
				if (exists($rib_in{$in}{$pref}{$exit})){
					# Resetting the traffic distribution and igp cost found by applying the tweaking
					# contained in the individual.
					my %anothertmptrafdistri = ();
					my $anothertmpigpcost = $tmpigpcost;
					foreach my $tutu (keys %tmptrafdistri){
						$anothertmptrafdistri{$tutu} = $tmptrafdistri{$tutu};
					};
					
					#print "Ok, exit $exit seems reachable through rib-in\n";
					# Ok, exit point can be reached. Changing traffic anf IGP cost.
					my ($exit,$exitas) = split(/[\s]+/,$rib_in{$in}{$pref}{$exit});
					my @tempexit = split(/\./,$exit);
					my ($default,$defaultas) = split(/[\s]+/,$bestroutes{$in}{$pref});
					my @tempdefault = split(/\./,$default);
					if ($tempdefault[0] ne '62'){
						# Default exit point is external. Remove traffic.
						$anothertmptrafdistri{$default}-=$traffic{$ingress2source{$in}}{$pref}[$time];
						# Do not remove anything to the IGP cost since it is 0 for external peers.
					}
					else{
						# Default exit point is internal. Convert default exit point to peer.
						$anothertmptrafdistri{$egressplusas2peer{"$default $defaultas"}}-=$traffic{$ingress2source{$in}}{$pref}[$time];
						# Remove contribution to IGP cost due to internal default exit point.
						$anothertmpigpcost-=($igpcost{$in}{$default}*$traffic{$ingress2source{$in}}{$pref}[$time]);
					};
					# Now adding traffic to new exit point.
					if ($tempexit[0] ne '62'){
						# New exit point is external. Add traffic contribution.
						$anothertmptrafdistri{$exit}+=$traffic{$ingress2source{$in}}{$pref}[$time];
						# No IGP cost to add since it is 0 for external peers.
					}
					else{
						# New exit point is internal. Convert new exit point to peer.
						$anothertmptrafdistri{$egressplusas2peer{"$exit $exitas"}}+=$traffic{$ingress2source{$in}}{$pref}[$time];
						# Add contribution to IGP cost.
						$anothertmpigpcost+=($igpcost{$in}{$exit}*$traffic{$ingress2source{$in}}{$pref}[$time]);
					};
					
					# Computing objective value if we add this tweaking. It's quite simple, the traffic
					# objective is the max of the traffic, we try to balance it.
					my $firstmax = 0;
					foreach my $tutu (keys %anothertmptrafdistri){
						if ($anothertmptrafdistri{$tutu} > $firstmax){
							$firstmax = $anothertmptrafdistri{$tutu};
						};
					};
					# Traffic objective is rather the sum of squares of delta with respect to perfect
					# balance.
					my $tot = 0;
					foreach my $tutu (keys %anothertmptrafdistri){
						$tot+=$anothertmptrafdistri{$tutu};
					};
					my $forbruno = '';
					foreach my $tutu (keys %anothertmptrafdistri){
						$forbruno.= ($anothertmptrafdistri{$tutu}/$tot)." ";
					};					
					my $delta = 0;
					my $delta2 = 0;
					my $max = 0;
					my $best = '';					
					# Amount of traffic if balance was perfect.
					my $balance = ($tot/scalar(keys %anothertmptrafdistri));
					foreach my $tutu (keys %anothertmptrafdistri){
						$delta+= (abs($anothertmptrafdistri{$tutu}-$balance)/$balance)**1;
					};
					foreach my $tutu (keys %anothertmptrafdistri){
						$delta2+= (abs($anothertmptrafdistri{$tutu}-$balance)/$balance)**2;
					};
					foreach my $tutu (keys %peers){
						if ( abs( ($anothertmptrafdistri{$tutu}/$tot) - (1/scalar(keys %anothertmptrafdistri))) > $max){
							$max = abs( ($anothertmptrafdistri{$tutu}/$tot) - (1/scalar(keys %anothertmptrafdistri)));
							$best = $elem;
						};
					};
					# Computing the real objective in terms of $max.
					$max = 1 + ($anothertmptrafdistri{$tutu}/$tot) - (1/scalar(keys %anothertmptrafdistri));
					# IGP cost is taken as is.
					
					# Now we check that the solution is non-dominated, but it is a fast one, not a complete
					# since we do not have time to remove elements from the list, we only add if element is
					# non-dominated.
					my $domination_check = 0;
					$nondominatedsize = scalar(@nondominated);
					if ($nondominatedsize >= 1){
						my $nondom = 0;
						while ($nondom < (scalar(@nondominated)-1)){
							# Compare element $nondom with tried one.
							#if ((($firstmax/$tot) >= $nondominated[$nondom][5])&&($anothertmpigpcost >= $nondominated[$nondom][1])){
							if (($delta2 >= $nondominated[$nondom][0])&&($anothertmpigpcost >= $nondominated[$nondom][1])){
								# Modified individual is dominated, mark it as such.
								$domination_check = 1;
							};
							$nondom++;
						};
					};
					# If domination check did not found a better individual in all objectives, add this one.
					if ($domination_check == 0){
#					if ($firstmax < $oldmax){
						# Putting this individual in the population of accepted individuals. Note
						# that we cannot simply mark it since it will change in the next iterations of this loop
						# over the reachable exit points.
						$accepted_pop_size++;
						# Copying old elements of the individual.
						my $elem = 1;
						while ($elem <= $population[$individual][0]){
							$accepted[$accepted_pop_size-1][$elem] = $population[$individual][$elem];
							$elem++;
						};
						# Now copying the new tweak (ingress, prefix, exit point and AS).
						# The -1 in the last part of element $elem of the newly accepted individual means
						# that this tweak lifetime is non-application since we've just added it. We'll initialize
						# it to $ARGV[7] only when writing the best individual.
						$accepted[$accepted_pop_size-1][$elem] = "$in $pref $exit $exitas -1";
						$accepted[$accepted_pop_size-1][0] = $elem;
						# Adding objective values as among the non-dominated ones found.
						$nondominated[$accepted_pop_size-1][0] = $delta2;
						$nondominated[$accepted_pop_size-1][1] = $anothertmpigpcost/$totcost;
						$nondominated[$accepted_pop_size-1][2] = $accepted_pop_size-1;
						$nondominated[$accepted_pop_size-1][3] = $population[$individual][0]+1;
						$nondominated[$accepted_pop_size-1][4] = $delta;
						$nondominated[$accepted_pop_size-1][5] = $firstmax/$tot;
						$nondominated[$accepted_pop_size-1][6] = $forbruno;
					};
				}
				else{
					#print "Exit $exit seems unreachable through rib-in\n";
				};
			};	
		}
		else{
			# Not ok, we have not been able to find a feasible pair within 5 tries.
			#print "Oups, could not find feasible pair within 5 tries\n";
		};	
	};
}

##################################
# Non-domination check
##################################
# This function checks that the accepted individuals are actually non-dominated.
sub nondomination_check {
	foreach my $first (0..($accepted_pop_size-1)){
		# For each element we check that it is not dominated by another individual (remove it),
		# or that it does not dominate other individials (remove them).
		if ($nondominated[$first][2] != -1){
			foreach my $second (0..($accepted_pop_size-1)){
				if (($nondominated[$second][2] != -1)&&($nondominated[$first][2] != -1)){
					# Testing if first is not dominated in which case we remove it.
					if (($nondominated[$first][0] > $nondominated[$second][0])&&($nondominated[$first][1] > $nondominated[$second][1])){
						# First is dominated by second. Remove first element.
						$nondominated[$first][0] = 0;
						$nondominated[$first][1] = 0;
						$nondominated[$first][2] = -1;
						$nondominated[$first][3] = 0;
						$nondominated[$first][4] = 0;
						$nondominated[$first][5] = 0;
						$nondominated[$first][6] = 0;
					}
					# Testing if second is not dominated in which case we remove it.
					elsif (($nondominated[$first][0] < $nondominated[$second][0])&&($nondominated[$first][1] < $nondominated[$second][1])){
						# First dominates second. Remove second element.
						$nondominated[$second][0] = 0;
						$nondominated[$second][1] = 0;
						$nondominated[$second][2] = -1;
						$nondominated[$second][3] = 0;
						$nondominated[$second][4] = 0;
						$nondominated[$second][5] = 0;
						$nondominated[$second][6] = 0;
					};
				};
			};
		};	
	};
}

##################################
# Computing the crowding distance
##################################
# This function computes the crowding distance for all individuals of the population. We take as input the @nondominated
# array and compute for each individual its crowding distance. Code comes from p.248 of book by Deb on MOP.
sub crowding_distance {

	my $time = $_[0];
	my $generation = $_[1];
	my $packedsize = 0;
	# First compute the number of non-dominated solutions.
	foreach $element (0..($accepted_pop_size-1)){
		if ($nondominated[$element][2] != -1){
			# Element is ok. Copy it into packednondominated.
			$packednondominated[$packedsize][0] = $nondominated[$element][0];
			$packednondominated[$packedsize][1] = $nondominated[$element][1];
			$packednondominated[$packedsize][2] = $nondominated[$element][2];
			$packednondominated[$packedsize][3] = $nondominated[$element][3];
			$packednondominated[$packedsize][4] = $nondominated[$element][4];
			$packednondominated[$packedsize][5] = $nondominated[$element][5];
			$packednondominated[$packedsize][6] = $nondominated[$element][6];
			$packedsize++;
			#print "copied $packedsize non-dominated individuals\n";
		};
	};
	print "Number of different non-dominated individuals is $packedsize\n";

	# Now sort the solutions by increasing objective value for each objective.
	# Traffic first, then delay.
	@traffic_sorted = sort {$packednondominated[$a][0] <=> $packednondominated[$b][0]} (0..(scalar(@packednondominated)-1));
	@igp_sorted = sort {$packednondominated[$a][1] <=> $packednondominated[$b][1]} (0..(scalar(@packednondominated)-1));

	# Now we compute the distance for each objective.
	# We assign a large distance for boundary solutions, other solutions are assigned the difference in fitness
	# between the next and the previous solutions divided by the difference between the best and the worst solutions.
	$crowd_dist_traffic[0] = 999999999999;
	$crowd_dist_traffic[scalar(@packednondominated)-1] = 999999999999;
	$crowd_dist_igp[0] = 999999999999;
	$crowd_dist_igp[scalar(@packednondominated)-1] = 999999999999;
	# range for each objective
	my $max_traffic = $packednondominated[$traffic_sorted[scalar(@packednondominated)-1]][0] - $packednondominated[$traffic_sorted[0]][0];
	my $max_igp = $packednondominated[$igp_sorted[scalar(@packednondominated)-1]][1] - $packednondominated[$igp_sorted[0]][1];
	# To prevent division by 0
	if ($max_traffic == 0){
		$max_traffic = 1;
	};
	# To prevent division by 0
	if ($max_igp == 0){
		$max_igp = 1;
	};
	# For each non-dominated element, we compute the distance in terms of each objective with the elements around it.
	foreach $elem (1..(scalar(@packednondominated)-2)){
		$crowd_dist_traffic[$traffic_sorted[$elem]] = abs(($packednondominated[$traffic_sorted[$elem+1]][0]-$packednondominated[$traffic_sorted[$elem-1]][0])/$max_traffic);
		$crowd_dist_igp[$igp_sorted[$elem]] = abs(($packednondominated[$igp_sorted[$elem+1]][1]-$packednondominated[$igp_sorted[$elem-1]][1])/$max_igp);
	};

	# Now that we did compute the crowding distance for each objective, we need to add both objectives' distance for each
	# individual.
	foreach $elem (0..(scalar(@packednondominated)-1)){
		$crowding_distance[$elem] = $crowd_dist_traffic[$elem] + $crowd_dist_igp[$elem];
#		$crowding_distance[$elem] = $crowd_dist_traffic[$elem];	
	};

	# Write the non-dominated individuals
	open(OUT,">>objective-space-results")|| die "Could not open output file: $!\n";
	foreach $elem (0..(scalar(@packednondominated)-1)){
		# Writing in output file the number of BGP filters of each individual, as well as their
		# traffic and delay objective values.
		print OUT "$time $generation $elem $packednondominated[$elem][3] $packednondominated[$elem][0] $packednondominated[$elem][1] $packednondominated[$elem][4] $packednondominated[$elem][5] $packednondominated[$elem][6]\n";
	};
	close(OUT)|| die "Could not close output file: $!\n";
}

########################################################################
# Compare last solution with non-dominated ones for this time interval.
########################################################################
# This function compares the last solution with the current non-dominated front. The comparison is made with
# respect to the ingress - prefixes pairs tweaked.

sub compare_solutions {
	my $time = $_[0];
	my $generation = $_[1];	
	%common_tweaks = ();
	# Comparing the number of (ingress,prefix,exit) tuples that differ between
	# the last solution and the current solutions found at this iteration. We
	# maintain in a list the non-dominated solutions that differ the least
	# from the last one.
	foreach $elem (0..(scalar(@packednondominated)-1)){
		# Comparing this individual with the last one.
		# For each tweaking of the current, compare it with all those of the
		# last solution.
		#print "Working on element $elem\n";
		my $tweak = 1;
		while ($tweak <= $accepted[$packednondominated[$elem][2]][0]){
			# Compare this tweak to each element of the last solution.
			my $lasttweak = 1;
			while ($lasttweak <= $last_solution[$time-1][0]){
				# Since tweaks may differ only by their lifetime, we only compare the four first elements
				# of the tweak. So splitting the string and then glueing the first fourth elements.
				my @lastlist =  split(/[\s]+/,$last_solution[$time-1][$lasttweak]);
				my @acceptedlist =  split(/[\s]+/,$accepted[$packednondominated[$elem][2]][$tweak]);
				my $laststring = $acceptedlist[0]." ".$acceptedlist[1]." ".$acceptedlist[2]." ".$acceptedlist[3];
				my $acceptedstring = $lastlist[0]." ".$lastlist[1]." ".$lastlist[2]." ".$lastlist[3];				
				if ($laststring eq $acceptedstring){
					# Found a common tweak between the two.
					$common_tweaks{$elem}++;
				};
				$lasttweak++;
			};
			$tweak++;
		};
	};
	# Write distance in terms of BGP changes only if we are within the generation range (unless we do not need
	# to write it since it has already been done in the main loop of this time interval).
	if ($generation <= MAXGEN){
		# Ok, we have compared the current front with the last solution. Writing in output file
		# the number of common tweaks compared to last solution.
		open(OUT,">>parameter-space-results")|| die "Could not open output file: $!\n";
		foreach $elem (0..(scalar(@packednondominated)-1)){
			if (defined($common_tweaks{$elem})){
				print OUT "$time $generation $elem $common_tweaks{$elem} ",$last_solution[$time-1][0],"\n";
			}
			else{
				print OUT "$time $generation $elem 0 ",$last_solution[$time-1][0],"\n";		
			};
		};
		close(OUT)|| die "Could not close output file: $!\n";
	};
}

#####################################################
# Chooses the best solution for this time interval.
#####################################################
# This function chooses the solution that is supposed to be used for the next interval
# (if real traffic engineering was to be performed). This solution is the one among those
# of the non-dominated ones whose distance from the last solution is minimal in terms of
# the BGP tweaking to be performed.

sub choose_last {
	
	my $time = $_[0];
	# Order the last front by decreasing number of differences with respect to the
	# last solution.
	my @orderedtweaks = sort {$common_tweaks{$b} <=> $common_tweaks{$a}} (keys %common_tweaks);
	# Then choose the smallest element that has the smallest value of the IGP cost.
	my $min = 9999999999;# That much should be enough as a min :)
	my $winnner = -1;
	my $elem = 0;
	while ($elem < (scalar(@orderedtweaks)-1)){
		if ($common_tweaks{$elem} == $common_tweaks{$orderedtweaks{0}}){
			# Ok, element has the same number of tweaks as the one that has most of them.
			if ($nondominated[$orderedtweaks[$elem]][1] < $min){
				$min = $nondominated[$orderedtweaks[$elem]][1];
 				$winner = $orderedtweaks[$elem];
			};
		}
		else{
			last;
		};
		$elem++;
	};
	# If winner is not -1, it means we have the winner. In that case we write this element
	# into last_solution. But here we deal with the tabu list propery of the last element.
	# We initialize the tweaks that have a -1 lifetime since these are newly added tweaks.
	# We decrement other lifetimes by one, and check if the TTL is found to be 0 we remove
	# these tweaks (actually we do not add them).
	if ($winner != -1){
		# We do not know yet how many tweaks there will be at the end. So only after
		# dealt with all tweaks do we modify the zeroth element that tells how many tweaks
		# lie within the individual.
		# Now we write this best individual into the output file.
		open(OUT,">>best-solutions")|| die "Could not open output file: $!\n";		
		my $counter = 0; # counts the number of valid tweaks.
		# Iterating over accepted individual.
		my $elem = 1;
		print OUT "$time $accepted[$packednondominated[$winner][2]][0]:";
		while ($elem <= $accepted[$packednondominated[$winner][2]][0]){
			# Splitting the element to deal with lifetime.
			my @tweak = split(/[\s]+/,$accepted[$packednondominated[$winner][2]][$elem]);
			# Now checking lifetime.
			if ($tweak[4] == -1){
				print OUT " $accepted[$packednondominated[$winner][2]][$elem]";
				# Ok, new entry. Initialize its value to $LIFETIME. We check if lifetime to be put is not 0
				# unless we cannot even add this tweak.
				if ($LIFETIME > 0){
					$counter++;
					$last_solution[$time][$counter] = $tweak[0]." ".$tweak[1]." ".$tweak[2]." ".$tweak[3]." $LIFETIME";
				}
				else{
					# Cannot add this tweak.
				};
			}
			elsif($tweak[4] > 1){
				$counter++;
				# If lifetime is larger than 1, decrement it.
				$last_solution[$time][$counter] = $tweak[0]." ".$tweak[1]." ".$tweak[2]." ".$tweak[3]." ".($tweak[4]-1);
				print OUT " $accepted[$packednondominated[$winner][2]][$elem]";
			}
			elsif($tweak[4] == 1){
				# Lifetime after decrementing will be zero, so drop this tweak :) (but write it in the file though since
				# this tweak influenced the objective values).
				print OUT " $accepted[$packednondominated[$winner][2]][$elem]";
			}
			else{
				# This should never happen since no tweak can be in an accepted individual with
				# a lifetime of 0 but should not have been added during the last time interval.
				print "Tweaks with zero lifetime found: $accepted[$packednondominated[$winner][2]][$elem]\n";
				exit(-1);
			};
			$elem++;
		};
		$last_solution[$time][0] = $counter;
		print OUT " $packednondominated[$winner][6]\n";
		close(OUT)|| die "Could not close output file: $!\n";
	}
	else{
		print "Oups, we were not able to find a best solution for time $time\n";
	};
}

######################################
# Population selection and generation
######################################
# This function selects the individuals from the last non-dominated front and generates the new population. We perform indeed
# a tournament selection among the non-dominated individuals by taking into account the crowding distance. The individuals having
# the largest crowding distance win the tournament since it is more important to include them in the population than individuals
# whose crowding distance is smaller. The selection process could not be simpler: we randomly choose two individuals among the
# non-dominated ones, the one with the largest crowding distance wins and is added to the next population, until we have enough
# individuals in the population. Recall that since the non-dominated individuals are in the @accepted array, we can replace the
# content of the @population array without the least care since it has already been freed.
sub select_generate_pop {
	my $first;
	my $second;
	foreach $individual (0..(MAXPOP-1)){
		$first = random_int_in(0,(scalar(@packednondominated)-1));
		$second = random_int_in(0,(scalar(@packednondominated)-1));
		if ($crowding_distance[$first] > $crowding_distance[$second]){
			# First element wins the tournament, write it into individual.
			my $elem = 0;
			while ($elem <= $accepted[$packednondominated[$first][2]][0]){
				$population[$individual][$elem] = $accepted[$packednondominated[$first][2]][$elem];
				$elem++;
			};
		}
		else{
			# Second element wins the tournament
			my $elem = 0;
			while ($elem <= $accepted[$packednondominated[$second][2]][0]){
				$population[$individual][$elem] = $accepted[$packednondominated[$second][2]][$elem];
				$elem++;
			};
		};
	};
}

#####################
# Load BGP updates
#####################
# This subroutine takes as input the time (hour) for which the BGP updates need to be loaded
# and filters them to maintain only those that are relevant for the traffic.

sub load_updates {

	my $time = $_[0];
	# The beginning timestamp we need to take into account is the one of the RIB.
	# So we convert the dates we see in the files into UNIX time.
	if ($time > 0){
		$BASETIME+= 3600;
	};
	#print "Base Unix time is $BASETIME\n";
	# Now we can begin searching for update files that are within the right hour.
	# The right hour is between BASETIME and BASETIME+3600.
	%FILELIST = ();
	finddepth(\&filter_updates,$ARGV[4]);
	# So now we should have in FILELIST the list of update files to consider for this
	# time period. We sort the name of the files according to their UNIX timestamps.
	@updatefiles = sort {$a <=> $b} (keys %FILELIST);
	# Now we load these update files one at a time.
	foreach $file (@updatefiles){
		# Working on file $FILELIST{$file}.
		print "Loading update file $FILELIST{$file}\n";
		# Opening the update file and filtering these updates since we only need
		# to care about those for which we have traffic.
		read_updates($FILELIST{$file});
		# Letting CBGP run the simulation to let BGP converge with the updates we sent.
		$cbgp->send("sim run\n");
	};
}

############################
# Updates RIB after updates 
############################
# This subroutine updates the RIBs after having loaded the BGP updates.
sub update_rib {
	# For each prefix with traffic.
	my $router = '';
	print "Number of prefixes to be updated: ",scalar(keys %tobeupdated),"\n";
	foreach my $pref (keys %tobeupdated){
		# Finally we ask to CBGP to dump the RIB of each ingress router of GEANT.
		foreach $in (keys %ingresses){
			# Asking best route for this ingress router.
 			$cbgp->send("print \"\# $in RIB for prefix $pref\\n\" \n");
			$cbgp->send("bgp router $in show rib-in * $pref\n");
			$cbgp->send("print \"done\\n\"\n");
			# Finished sending commands to CBGP for this prefix. Now we read from the input pipe the output
			# of CBGP.
			while ($_= $cbgp->expect(1)){
				chomp;
				m/^done/ and last;
				# Working on line we read.
				# If line begins with "#" then it contains the IP of the GEANT router
				# whose RIB we read. Next line then contains the best route to reach
				# prefix $pref.
				#print $_;
				my @temp = split(/[\s]+/,$_);
				if ($temp[0] eq '#'){
					# Read next line, record which router it is the RIB.
					$router = $temp[1];
					$prefix = $temp[5];
				}
				elsif ($temp[0] eq '*>'){
					# Ok, checking that prefix is the same in the RIB
					if ($temp[1] ne $prefix){
						print "Prefix in RIB-in ($temp[1]) is different from expected one: $prefix\n";
					};
					if (($router ne '')&&($temp[2] ne '')){
						$bestroutes{$router}{$pref} = "$temp[2] $temp[5]";
						#print "best egress for $router to reach $pref is $temp[2] $temp[5]\n"; 
					}
					else{
						print "Found void exit router in update_rib ($router $temp[2])\n";
					};
				}
				elsif ($temp[0] eq '*'){
					if ($temp[1] ne $prefix){
						print "Prefix in RIB-in ($temp[1]) is different from expected one: $prefix\n";
					};
					my @toto = split(/\./,$temp[2]);
					if (($router ne '')&&($temp[2] ne '')){
						if ($toto[0] eq '62'){
							$rib_in{$router}{$pref}{$egressplusas2peer{"$temp[2] $temp[5]"}} = "$temp[2] $temp[5]";
							#print "egress for $router to reach $pref is ",$egressplusas2peer{"$temp[2] $temp[5]"},"\n"; 
						}
						else{
							$rib_in{$router}{$pref}{$temp[2]} = "$temp[2] $temp[5]";
							#print "egress for $router to reach $pref is $temp[2]\n";
						};
					}
					else{
						print "Found void exit router in update_rib ($router $temp[2])\n";
					};
				}
				else{
					print "Oups, this is a strange line received from cbgp: $_";
				};
			};
		};
		# No more line to read from CBGP.
		#print "No more line to read from CBGP, going to next prefix\n";	
	};
}


#####################
# Read BGP updates
#####################
# This subroutine takes as input a file that contains BGP updates.

sub read_updates {
	my $pid;
	if (not defined($pid = open(IN, "-|"))) {
		die "can't fork: $!";
	}
	if ($pid) {
		# parent process - do nothing
	}
	else {
		# child process
		system("zcat $_[0] ");
		exit 0;
	};
	while (defined($CURRENTLINE=<IN>)){
		chomp($CURRENTLINE);
		my @route = split(/\|/,$CURRENTLINE);
		# Accepting only routes that concern the destination prefixes.
		if ((exists($trafprefix{$route[5]}))&&(defined($trafprefix{$route[5]}))&&($route[5] ne '')){
			# So there is some traffic towards this destination prefix.
			# But still, we check that the next hop is really among those
			# that we expect (GlobalCrossing or Telia)
			
			# So now things are a little more complex depending on whether we see an announce or
			# a withdraw.
			if ($route[2] eq 'W'){
 				#print "$CURRENTLINE\n";
				# Marking prefix as having potentially changed (for incremental RIB update).
				$tobeupdated{$route[5]} = 1;
				# Cleaning old elements of router's RIB's and Rib-in's for prefix.
				foreach $in (keys %ingresses){
					# Removing best route we recorded for this prefix.
					if (exists($besroutes{$in}{$route[5]})){
						undef $besroutes{$in}{$route[5]};
					};
					if (exists($rib_in{$in}{$route[5]})){
						undef $rib_in{$in}{$route[5]};
					};
				};

				# Update is a withdraw, so we do not know from which peer it comes. Hence announcing
				# the withdraw as if each peer router connected to the GEANT router was to announce
				# the update.
				my $test = 0;
				foreach my $peerrouter (keys %peers){
					if (exists($externalpeerings{"$route[3] $peerrouter"})){
						# Found a peer of the local GEANT router.
						$cbgp->send("bgp router $route[3] peer $peerrouter recv ".
						'"BGP4|0|'.$route[2].'|'.$route[3].'|'.$route[4].'|'.$route[5].'"'."\n");
						$test = 1;
					};
				};
				if ($test == 0){
					print "Oups, could not find an external peer for this update message: $CURRENTLINE\n";
				};
			}
			else{
				my @ALIST = split(/\./,$route[8]);
				if (($ALIST[0] != 208)&&($ALIST[0] != 213)){
					# This should never happen, something is wrong in this route
					# since the next hop does not seem to be one of the peers
					# we expect.
					print "Oups \! Found next hop not among those we expect in updates: $route[8]\n";
				}
				else{
	 				#print "$CURRENTLINE\n";
					# Marking prefix as having potentially changed (for incremental RIB update).
					$tobeupdated{$route[5]} = 1;
					# Cleaning old elements of router's RIB's and Rib-in's for prefix.
					foreach $in (keys %ingresses){
						# Removing best route we recorded for this prefix.
						if (exists($besroutes{$in}{$route[5]})){
							undef $besroutes{$in}{$route[5]};
						};
						if (exists($rib_in{$in}{$route[5]})){
							undef $rib_in{$in}{$route[5]};
						};
					};
					# Everything is fine, prefix has traffic and next hop is known.
					# We record this route since we'll inject it into CBGP when we'll
					# have seen all the routes towards those prefixes we're interested
					# in.
					# Injecting update into CBGP.
					$cbgp->send("bgp router $route[3] peer $route[8] recv ".
					'"BGP4|0|'.$route[2].'|'.$route[3].'|'.$route[4].'|'.$route[5].'|'.$route[6].'|IGP|'.$route[8].'|'.$route[9].'|'.$route[10].'||NAG||"'."\n");
				};
			};
		}
		else{
			#print $CURRENTLINE;
		};
	};# end while
}

########################
# Filters update files
########################

sub filter_updates {
	# Checking that this is a real file.
	if (-f $File::Find::name){
		# Then checking that this is an update file with a correct name.
		my @FILENAME = split(/\//,$File::Find::name);
		my @DATE = split(/\./,$FILENAME[scalar(@FILENAME)-1]);
		# Last element of DATE should be 'gz', first should be 'updates.
		if (($DATE[scalar(@DATE)-1] eq 'gz')&&($DATE[0] eq 'updates')){
			# Converting this file's date info into Unix seconds.
			my $month = substr($DATE[scalar(@DATE)-3],4,2);
			my $day = substr($DATE[scalar(@DATE)-3],6,2);
			my $year = substr($DATE[scalar(@DATE)-3],0,4);
			my $hour = substr($DATE[scalar(@DATE)-2],0,2);
			my $minutes = substr($DATE[scalar(@DATE)-2],2,2);
			# Base UNIX time is as follows.
			$FILETIME = &Date_SecsSince1970($month,$day,$year,$hour,$minutes,0);
			# Now checking if the hour indicated in the file is within one hour
			# from the BASETIME.
			if (($FILETIME >= $BASETIME)&&(($FILETIME - $BASETIME) < 3600)){
				# Ok, file should be included in the list.
				$FILELIST{$FILETIME} = $File::Find::name;
				#print "Added file $FILELIST{$FILETIME} to the list of update files ($FILETIME  $BASETIME)\n";
			};
		};
	};
}

###################
# Randint function
###################

sub random_int_in ($$) {
    my($min, $max) = @_;
    # Assumes that the two arguments are integers themselves!
    return $min if $min == $max;
    ($min, $max) = ($max, $min)  if  $min > $max;
    return $min + int(rand(1 + $max - $min));
}  
